@@ -3,28 +3,26 @@
 //!
 //! - A `meta` file with metadata about the object.
 //! - One or more binary data chunks.
-use std::{fmt::Debug, str::FromStr, string::FromUtf8Error, time::Instant};
+use std::{fmt::Debug, iter, str::FromStr, string::FromUtf8Error, sync::Arc, time::Instant};
 
-use crate::{bucket::BucketName, object::meta::get_meta, Context};
+use crate::{bucket::BucketName, object::meta::get, Context};
 use bytes::{Bytes, BytesMut};
 use chrono::Utc;
+use derive_more::{AsRef, Deref, DerefMut, Display};
 use futures_util::{
-    future,
     stream::{self},
-    Future, Stream, StreamExt, TryStreamExt,
+    AsyncBufRead, AsyncReadExt, Stream, StreamExt, TryStreamExt,
 };
 
 use jotta_fs::{
     files::{AllocReq, ConflictHandler, UploadRes},
     path::{PathOnDevice, UserScopedPath},
-    ByteRange,
+    range::{ByteRange, ClosedByteRange, OpenByteRange},
 };
-use mime::Mime;
 use serde::{Deserialize, Serialize};
-use tokio::io::{AsyncBufRead, AsyncReadExt};
 use tracing::{debug, error, instrument, trace, warn};
 
-use self::meta::{set_meta, ObjectMeta};
+use self::meta::{set_raw, Meta, Patch};
 
 pub mod meta;
 
@@ -48,7 +46,20 @@ pub const CHUNK_SIZE: usize = 1 << 20;
 /// assert!(ObjectName::from_str("hello\nworld").is_err());
 /// assert!(ObjectName::from_str("bye\r\nlword").is_err());
 /// ```
-#[derive(Debug, Serialize, Deserialize)]
+#[derive(
+    Debug,
+    Serialize,
+    Deserialize,
+    Clone,
+    PartialEq,
+    Eq,
+    PartialOrd,
+    Ord,
+    Deref,
+    DerefMut,
+    AsRef,
+    Display,
+)]
 #[allow(clippy::module_name_repetitions)]
 pub struct ObjectName(String);
 
@@ -130,10 +141,7 @@ pub enum InvalidObjectName {
 ///
 /// Returns an error if there is no bucket with the specified name.
 #[instrument(skip(ctx))]
-pub async fn list_objects(
-    ctx: &Context,
-    bucket: &BucketName<'_>,
-) -> crate::Result<Vec<ObjectName>> {
+pub async fn list_objects(ctx: &Context, bucket: &BucketName) -> crate::Result<Vec<ObjectName>> {
     let folders = ctx
         .fs
         .index(&UserScopedPath(format!(
@@ -150,30 +158,33 @@ pub async fn list_objects(
         .collect::<crate::Result<Vec<_>>>()
 }
 
-/// Create an object. This does not upload any actual blobs, only metadata.
+/// Create an object. This does not upload any actual binary data, only metadata.
 #[instrument(skip(ctx))]
 pub async fn create_object(
     ctx: &Context,
-    bucket: &BucketName<'_>,
+    bucket: &BucketName,
     name: &ObjectName,
-    content_type: Option<Mime>,
-) -> crate::Result<()> {
+    meta: Patch,
+) -> crate::Result<Meta> {
     let now = Utc::now();
 
-    let meta = ObjectMeta {
+    let meta = Meta {
         size: 0,
         created: now,
         updated: now,
-        content_type: content_type.unwrap_or(mime::APPLICATION_OCTET_STREAM),
+        content_type: meta.content_type.unwrap_or_default(),
+        cache_control: meta.cache_control.unwrap_or_default(),
     };
 
-    set_meta(ctx, bucket, name, &meta, ConflictHandler::RejectConflicts).await
+    set_raw(ctx, bucket, name, &meta, ConflictHandler::RejectConflicts).await?;
+
+    Ok(meta)
 }
 
 #[instrument(level = "trace", skip(ctx, bucket, name, body))]
 async fn upload_chunk(
     ctx: &Context,
-    bucket: &BucketName<'_>,
+    bucket: &BucketName,
     name: &ObjectName,
     index: u32,
     body: Bytes, // there is no point accepting a stream since a checksum needs to be calculated prior to allocation anyway
@@ -207,7 +218,7 @@ async fn upload_chunk(
 
 async fn get_complete_chunk<R: AsyncBufRead + Unpin>(
     ctx: &Context,
-    bucket: &BucketName<'_>,
+    bucket: &BucketName,
     name: &ObjectName,
     mut cursor: usize,
     chunk_no: u32,
@@ -225,7 +236,7 @@ async fn get_complete_chunk<R: AsyncBufRead + Unpin>(
             .fs
             .file_to_bytes(
                 chunk_path,
-                ByteRange::try_from_bounds(0..(cursor as u64)).unwrap(),
+                ClosedByteRange::new_to_including(cursor as u64 - 1),
             )
             .await?;
 
@@ -257,14 +268,17 @@ async fn get_complete_chunk<R: AsyncBufRead + Unpin>(
         // If the case is the latter, we need to download the tail of this chunk in order not to
         // accidentally truncate the file.
 
-        let tail = ctx
+        println!("we need tail");
+
+        let tail = match ctx
             .fs
-            .file_to_bytes(
-                chunk_path,
-                ByteRange::try_from_bounds(cursor as u64..).unwrap(),
-            )
+            .file_to_bytes(chunk_path, OpenByteRange::new(cursor as u64))
             .await
-            .unwrap_or_default(); // TODO: actual error handling
+        {
+            Ok(bytes) => bytes,
+            Err(jotta_fs::Error::NoSuchFileOrFolder) => Bytes::new(), // no tail was found. no worries
+            Err(e) => return Err(e.into()),
+        };
 
         buf.extend_from_slice(&tail);
     }
@@ -277,12 +291,12 @@ async fn get_complete_chunk<R: AsyncBufRead + Unpin>(
 #[instrument(skip(ctx, file))]
 pub async fn upload_range<R: AsyncBufRead + Unpin>(
     ctx: &Context,
-    bucket: &BucketName<'_>,
+    bucket: &BucketName,
     name: &ObjectName,
     offset: u64,
     file: R,
     num_connections: usize,
-) -> crate::Result<()> {
+) -> crate::Result<Meta> {
     let before = Instant::now();
 
     let chunks = stream::try_unfold((file, offset), move |(mut file, pos)| async move {
@@ -322,71 +336,90 @@ pub async fn upload_range<R: AsyncBufRead + Unpin>(
         bytes_per_second * 8.0 / 1_000_000.0
     );
 
-    let meta = get_meta(ctx, bucket, name).await?;
+    let meta = get(ctx, bucket, name).await?;
 
-    let meta = ObjectMeta {
+    let meta = Meta {
         size: meta.size.max(bytes_uploaded + offset),
         updated: Utc::now(),
         ..meta
     };
 
-    set_meta(ctx, bucket, name, &meta, ConflictHandler::CreateNewRevision).await
+    set_raw(ctx, bucket, name, &meta, ConflictHandler::CreateNewRevision).await?;
+
+    Ok(meta)
+}
+
+fn aligned_chunked_byte_range(
+    range: impl ByteRange,
+) -> impl Iterator<Item = (u32, ClosedByteRange)> {
+    let mut pos = range.start();
+
+    iter::from_fn(move || {
+        #[allow(clippy::cast_possible_truncation)]
+        let chunk_no = (pos / (CHUNK_SIZE as u64)) as u32;
+        let chunk_start = pos % (CHUNK_SIZE as u64);
+
+        let chunk_end = (range.end().unwrap_or(u64::MAX) - pos).min(CHUNK_SIZE as _);
+
+        if chunk_end == 0 {
+            return None;
+        }
+
+        let chunk = ClosedByteRange::try_from_bounds(chunk_start, chunk_end).unwrap();
+
+        pos += chunk_end - chunk_start;
+
+        Some((chunk_no, chunk))
+    })
 }
 
 /// Open a stream to an object.
 ///
 /// **The integrity of the data is not checked by this function.**
+///
+/// # Errors
+///
+/// The stream will eventually return an error if `range` is infinite,
+/// since there won't be enough chunks in the cloud to satisfy the
+/// range.
 #[instrument(skip(ctx))]
 #[allow(clippy::manual_async_fn)] // lifetimes don't allow async syntax
-pub fn open_range<'a>(
-    ctx: &'a Context,
-    bucket: &'a BucketName<'a>,
-    name: &'a ObjectName,
+pub fn stream_range<'a>(
+    ctx: Arc<Context>,
+    bucket: BucketName,
+    name: ObjectName,
+    range: ClosedByteRange,
     num_connections: usize,
-) -> impl Future<Output = crate::Result<(ObjectMeta, impl Stream<Item = crate::Result<Bytes>> + 'a)>> + 'a
-{
-    async move {
-        let meta = get_meta(ctx, bucket, name).await?;
+) -> impl Stream<Item = crate::Result<Bytes>> + 'a {
+    stream::iter(aligned_chunked_byte_range(range))
+        .map(move |(chunk_no, range)| {
+            let ctx = ctx.clone();
+            let bucket = bucket.clone();
+            let name = name.clone();
 
-        let futures = stream::iter(0..)
-            .map(move |chunk_no| async move {
-                // let ctx = ctx.clone();
-
-                match ctx
-                    .fs
+            async move {
+                ctx.fs
                     .file_to_bytes(
                         &UserScopedPath(format!(
-                            "{}/{bucket}/{}",
+                            "{}/{}/{}",
                             ctx.user_scoped_root(),
-                            name.chunk_path(chunk_no)
+                            &bucket,
+                            name.clone().chunk_path(chunk_no)
                         )),
-                        ByteRange::full(),
+                        range,
                     )
                     .await
-                {
-                    // Ok(s) => Some(s),
-                    Ok(b) => Some(Ok(b)),
-                    Err(e) => {
-                        warn!("encountered an error while reading chunk: {:?}", e);
-                        None
-                    }
-                }
-            })
-            .buffered(num_connections);
-
-        let stream = futures
-            .take_while(|d| future::ready(d.is_some()))
-            .map(Option::unwrap);
-
-        Ok((meta, stream))
-    }
+            }
+        })
+        .buffered(num_connections)
+        .map_err(Into::into)
 }
 
 /// Delete an object.
 #[instrument(skip(ctx))]
 pub async fn delete_object(
     ctx: &Context,
-    bucket: &BucketName<'_>,
+    bucket: &BucketName,
     name: &ObjectName,
 ) -> crate::Result<()> {
     let _res = ctx
@@ -400,3 +433,50 @@ pub async fn delete_object(
 
     Ok(())
 }
+
+#[cfg(test)]
+mod tests {
+    use jotta_fs::range::{ClosedByteRange, OpenByteRange};
+
+    use crate::object::{aligned_chunked_byte_range, CHUNK_SIZE};
+
+    #[test]
+    fn create_aligned_chunks() {
+        let mut iter = aligned_chunked_byte_range(OpenByteRange::full());
+
+        assert_eq!(
+            iter.next().unwrap(),
+            (0, ClosedByteRange::new_to_including(CHUNK_SIZE as _))
+        );
+        assert_eq!(
+            iter.next().unwrap(),
+            (1, ClosedByteRange::new_to_including(CHUNK_SIZE as _))
+        );
+        assert_eq!(
+            iter.next().unwrap(),
+            (2, ClosedByteRange::new_to_including(CHUNK_SIZE as _))
+        );
+
+        assert_eq!(
+            aligned_chunked_byte_range(ClosedByteRange::try_from(40..=2_500_000).unwrap())
+                .collect::<Vec<_>>(),
+            vec![
+                (0, ClosedByteRange::try_from_bounds(40, 1_048_576).unwrap()),
+                (1, ClosedByteRange::new_to_including(1_048_576)),
+                (2, ClosedByteRange::new_to_including(402_848))
+            ]
+        );
+
+        assert_eq!(
+            aligned_chunked_byte_range(ClosedByteRange::try_from(69_420_000..=71_000_000).unwrap())
+                .collect::<Vec<_>>(),
+            vec![
+                (
+                    66,
+                    ClosedByteRange::try_from_bounds(213_984, 1_048_576).unwrap()
+                ),
+                (67, ClosedByteRange::new_to_including(745_408))
+            ]
+        );
+    }
+}
